import torch
import torch.nn as nn
from mmcv.runner import BaseModule
from ..builder import NECKS
from typing import Optional, Union


class ConvNormAct(nn.Module):
    def __init__(self,
                 in_channel: int, out_channel: int, kernel_size: int = 3, stride=1, padding=None, groups: int = 1,
                 use_norm: Optional[bool] = False, bias=False, use_act: Optional[bool] = False,
                 dilation: int = 1, momentum: int = 0.1, out_size: tuple = (224, 224)):
        super(ConvNormAct, self).__init__()

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(stride, int):
            stride = (stride, stride)
        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        if padding is None:
            padding = (int((kernel_size[0] - 1) / 2) * dilation[0], int((kernel_size[1] - 1) / 2) * dilation[1])

        if use_norm:
            bias = False

        self.conv = torch.nn.Conv2d(in_channel, out_channel, kernel_size, stride, padding,
                                    dilation=dilation, groups=groups, bias=bias)

        self.norm_layer = nn.BatchNorm2d(out_channel, momentum=momentum) if use_norm else None
        self.act = nn.SiLU(inplace=True) if use_act else None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        if self.norm_layer is not None:
            x = self.norm_layer(x)
        if self.act is not None:
            x = self.act(x)
        return x


class IRB(nn.Module):
    """
    Inverted residual block (MobileNetv2): https://arxiv.org/abs/1801.04381
    """

    def __init__(self,
                 in_channel, out_channel,
                 expand_ratio: Union[int, float] = 2,
                 ) -> None:
        super(IRB, self).__init__()

        expansion_channel = int(round(in_channel * expand_ratio))

        self.conv_1x1_in = ConvNormAct(in_channel=in_channel, out_channel=expansion_channel, kernel_size=1,
                                       use_act=True, use_norm=True, )

        self.dw_3x3 = ConvNormAct(in_channel=expansion_channel, out_channel=expansion_channel, kernel_size=3,
                                  groups=expansion_channel, use_act=True, use_norm=True, )

        self.conv_1x1_out = ConvNormAct(in_channel=expansion_channel, out_channel=out_channel, kernel_size=1,
                                        use_act=False, use_norm=True, )

    def forward(self, x):
        x = self.conv_1x1_in(x)
        x = x + self.dw_3x3(x)
        return self.conv_1x1_out(x)


class SeparableConv(nn.Module):
    """
    This layer defines Depth-wise separable convolution, introduced in Xception
        https://arxiv.org/abs/1610.02357
    """

    def __init__(self,
                 in_channel: int,
                 out_channel: int,
                 padding,
                 kernel_size=3,
                 stride: Optional[int or tuple] = 1,
                 dilation: Optional[int or tuple] = 1,
                 use_norm: Optional[bool] = True,
                 use_act: Optional[bool] = True,
                 bias: Optional[bool] = False, padding_mode: Optional[str] = 'zeros',
                 *args, **kwargs):
        super(SeparableConv, self).__init__()
        self.dw_conv = ConvNormAct(in_channel=in_channel, out_channel=in_channel, kernel_size=kernel_size,
                                   stride=stride, padding=padding, dilation=dilation, groups=in_channel, bias=False,
                                   use_norm=True, use_act=False)
        self.pw_conv = ConvNormAct(in_channel=in_channel, out_channel=out_channel, kernel_size=1,
                                   stride=1, dilation=1, groups=1, bias=bias, use_norm=use_norm, use_act=use_act)

    def forward(self, x):
        x = self.dw_conv(x)
        x = self.pw_conv(x)
        return x


@NECKS.register_module()
class LcTNeck(BaseModule):
    """Extra layers of SSD backbone to generate multi-scale feature maps.

    Args:
        in_channels (Sequence[int]): Number of input channels per scale.
        out_channels (Sequence[int]): Number of output channels per scale.
        level_strides (Sequence[int]): Stride of 3x3 conv per level.
        level_paddings (Sequence[int]): Padding size of 3x3 conv per level.
        l2_norm_scale (float|None): L2 normalization layer init scale.
            If None, not use L2 normalization on the first input feature.
        last_kernel_size (int): Kernel size of the last conv layer.
            Default: 3.
        use_depthwise (bool): Whether to use DepthwiseSeparableConv.
            Default: False.
            Default: None.
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 project_ratio=2,
                 level_strides=(2, 1, 1),
                 level_paddings=(1, 0, 0),
                 l2_norm_scale=20.,
                 last_kernel_size=3,
                 use_depthwise=False,
                 init_cfg=[
                     dict(
                         type='Xavier', distribution='uniform',
                         layer='Conv2d'),
                     dict(type='Constant', val=1, layer='BatchNorm2d'),
                 ]):
        super(LcTNeck, self).__init__(init_cfg)
        assert len(out_channels) > len(in_channels)
        assert len(out_channels) - len(in_channels) == len(level_strides)
        assert len(level_strides) == len(level_paddings)

        # projection layers
        pro_layers = []
        for i in range(len(in_channels)):
            pro_layers.append(ConvNormAct(in_channel=in_channels[i], out_channel=out_channels[i], kernel_size=1,
                                          stride=1, padding=0, use_norm=True, use_act=True))
        self.pro_layers = nn.Sequential(*pro_layers)

        # extra layers
        extra_layers = []
        self.in_channel = in_channels[-1]
        extra_channels = out_channels[len(in_channels):]
        for i, (stride, padding) in enumerate(zip(level_strides, level_paddings)):
            extra_layers.append(SeparableConv(in_channel=self.in_channel, out_channel=extra_channels[i], stride=stride,
                                              padding=padding))
            self.in_channel = extra_channels[i]
        self.extra_layers = nn.Sequential(*extra_layers)

    def forward(self, inputs):

        outs = []

        for i, layer in enumerate(self.pro_layers):
            feat = layer(inputs[i])
            outs.append(feat)

        feat = inputs[-1]
        for layer in self.extra_layers:
            feat = layer(feat)
            outs.append(feat)
        return tuple(outs)

# class L2Norm(nn.Module):
#
#     def __init__(self, n_dims, scale=20., eps=1e-10):
#         """L2 normalization layer.
#
#         Args:
#             n_dims (int): Number of dimensions to be normalized
#             scale (float, optional): Defaults to 20..
#             eps (float, optional): Used to avoid division by zero.
#                 Defaults to 1e-10.
#         """
#         super(L2Norm, self).__init__()
#         self.n_dims = n_dims
#         self.weight = nn.Parameter(torch.Tensor(self.n_dims))
#         self.eps = eps
#         self.scale = scale
#
#     def forward(self, x):
#         """Forward function."""
#         # normalization layer convert to FP32 in FP16 training
#         x_float = x.float()
#         norm = x_float.pow(2).sum(1, keepdim=True).sqrt() + self.eps
#         return (self.weight[None, :, None, None].float().expand_as(x_float) *
#                 x_float / norm).type_as(x)
