from typing import Optional, Union

import torch
import torch.nn as nn
from mmcv.cnn import ConvModule

from mmseg.ops import resize
from ..builder import HEADS
from .decode_head import BaseDecodeHead


class ConvNormAct(nn.Module):
    def __init__(self,
                 in_channel: int, out_channel: int, kernel_size: int = 3, stride=1, groups: int = 1,
                 use_norm: Optional[bool] = False, bias=False, use_act: Optional[bool] = False,
                 dilation: int = 1, norm_type='BN'):
        super(ConvNormAct, self).__init__()

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(stride, int):
            stride = (stride, stride)
        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        padding = (int((kernel_size[0] - 1) / 2) * dilation[0], int((kernel_size[1] - 1) / 2) * dilation[1])

        if use_norm:
            bias = False

        self.conv = torch.nn.Conv2d(in_channel, out_channel, kernel_size, stride, padding,
                                    dilation=dilation, groups=groups, bias=bias)

        # self.norm_layer = nn.BatchNorm2d(out_channel, momentum=momentum) if use_norm else None

        self.norm_layer = nn.BatchNorm2d(out_channel) if use_norm else None
        # else:
        #     self.norm_layer = LayerNorm(out_channel, data_format="channels_first") if use_norm else None

        self.act = nn.SiLU(inplace=True) if use_act else None
        # self.act = nn.GELU() if use_act else None
        # self.act = nn.ReLU(inplace=True) if use_act else None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        if self.norm_layer is not None:
            x = self.norm_layer(x)
        if self.act is not None:
            x = self.act(x)
        return x


class IRB(nn.Module):
    """
    Inverted residual block (MobileNetv2): https://arxiv.org/abs/1801.04381
    """

    def __init__(self,
                 in_channel: int,
                 out_channel: int,
                 dilation=1,
                 expand_ratio: Union[int, float] = 2,
                 use_dw=True,
                 ) -> None:
        super(IRB, self).__init__()

        expansion_channel = int(round(in_channel * expand_ratio))

        self.conv_1x1_in = ConvNormAct(in_channel=in_channel, out_channel=expansion_channel, kernel_size=1,
                                       use_act=True, use_norm=True, )

        if use_dw:
            self.dw_3x3 = ConvNormAct(in_channel=expansion_channel, out_channel=expansion_channel, kernel_size=3,
                                      groups=expansion_channel, dilation=dilation,
                                      use_act=True, use_norm=True, )
        else:
            self.dw_3x3 = None

        self.conv_1x1_out = ConvNormAct(in_channel=expansion_channel, out_channel=out_channel, kernel_size=1,
                                        use_act=False, use_norm=True, )

    def forward(self, x):
        x = self.conv_1x1_in(x)
        if self.dw_3x3 is not None:
            x = x + self.dw_3x3(x)
        return self.conv_1x1_out(x)


class InvertedResidual(nn.Module):
    """
    Inverted residual block (MobileNetv2): https://arxiv.org/abs/1801.04381
    """

    def __init__(self,
                 in_channel: int,
                 out_channel: int,
                 stride: int,
                 expand_ratio: Union[int, float] = 4,
                 kernel_size=3,
                 dilation: int = 1,
                 use_se=False,
                 out_size: tuple = (224, 224),
                 norm_type="BN"
                 ) -> None:
        assert stride in [1, 2]
        super(InvertedResidual, self).__init__()
        self.stride = stride
        int_size = (int(out_size[0] * stride), int(out_size[1] * stride))

        expansion_channel = int(round(in_channel * expand_ratio))
        self.use_res_connect = self.stride == 1 and in_channel == out_channel

        block = nn.Sequential()
        if expand_ratio != 1:
            block.add_module(name="exp_1x1",
                             module=ConvNormAct(in_channel=in_channel, out_channel=expansion_channel, kernel_size=1,
                                                use_act=True, use_norm=True, norm_type=norm_type))
        block.add_module(
            name="conv_3x3",
            module=ConvNormAct(in_channel=expansion_channel, out_channel=expansion_channel, stride=stride,
                               kernel_size=kernel_size, groups=expansion_channel, use_act=True, use_norm=True,
                               dilation=dilation, norm_type=norm_type))

        block.add_module(name="red_1x1",
                         module=ConvNormAct(in_channel=expansion_channel, out_channel=out_channel, kernel_size=1,
                                            use_act=False, use_norm=True, norm_type=norm_type))

        self.block = block

    def forward(self, x):
        if self.use_res_connect:
            return x + self.block(x)
        else:
            return self.block(x)


class ASPPModule(nn.ModuleList):
    """Atrous Spatial Pyramid Pooling (ASPP) Module.

    Args:
        dilations (tuple[int]): Dilation rate of each layer.
        in_channels (int): Input channels.
        channels (int): Channels after modules, before conv_seg.
        conv_cfg (dict|None): Config of conv layers.
        norm_cfg (dict|None): Config of norm layers.
        act_cfg (dict): Config of activation layers.
    """

    def __init__(self, dilations, in_channels, channels, conv_cfg, norm_cfg,
                 act_cfg, aspp_ratio=1.5, ):
        super(ASPPModule, self).__init__()
        self.dilations = dilations
        self.in_channels = in_channels
        self.channels = channels
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        for dilation in dilations:
            # self.append(ConvNormAct(in_channel=self.in_channels,
            #                         out_channel=self.channels,
            #                         dilation=dilation,
            #                         use_norm=True,
            #                         use_act=True,
            #                         ))
            self.append(InvertedResidual(in_channel=self.in_channels,
                                         out_channel=self.channels,
                                         stride=1,
                                         dilation=dilation,
                                         expand_ratio=aspp_ratio,
                                         ))

            # self.append(
            #     ConvModule(
            #         self.in_channels,
            #         self.channels,
            #         1 if dilation == 1 else 3,
            #         dilation=dilation,
            #         padding=0 if dilation == 1 else dilation,
            #         conv_cfg=self.conv_cfg,
            #         norm_cfg=self.norm_cfg,
            #         act_cfg=self.act_cfg))

    def forward(self, x):
        """Forward function."""
        aspp_outs = []
        for aspp_module in self:
            aspp_outs.append(aspp_module(x))

        return aspp_outs


@HEADS.register_module()
class LcTASPPHead(BaseDecodeHead):
    """Rethinking Atrous Convolution for Semantic Image Segmentation.

    This head is the implementation of `DeepLabV3
    <https://arxiv.org/abs/1706.05587>`_.

    Args:
        dilations (tuple[int]): Dilation rates for ASPP module.
            Default: (1, 6, 12, 18).
    """

    def __init__(self, dilations=(1, 6, 12, 18), bottleneck_ratio=0.25, aspp_ratio=2, neck_dropout=0, **kwargs):
        super(LcTASPPHead, self).__init__(**kwargs)
        assert isinstance(dilations, (list, tuple))
        self.dilations = dilations
        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels,
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg))
        self.aspp_modules = ASPPModule(
            dilations,
            self.in_channels,
            self.channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            aspp_ratio=aspp_ratio, )
        self.bottleneck = IRB(in_channel=(len(dilations) + 1) * self.channels,
                              out_channel=self.channels,
                              expand_ratio=bottleneck_ratio,
                              )
        self.bottleneck_dropout = nn.Dropout(p=neck_dropout, inplace=True)

    def forward(self, inputs):
        """Forward function."""
        x = self._transform_inputs(inputs)
        aspp_outs = [
            resize(
                self.image_pool(x),
                size=x.size()[2:],
                mode='bilinear',
                align_corners=self.align_corners)
        ]
        aspp_outs.extend(self.aspp_modules(x))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        output = self.bottleneck_dropout(self.bottleneck(aspp_outs))
        output = self.cls_seg(output)
        return output
