from typing import Optional, Union

import torch
import torch.nn as nn
from mmcv.cnn import ConvModule

from ..builder import HEADS
from .decode_head import BaseDecodeHead


class ConvNormAct(nn.Module):
    def __init__(self,
                 in_channel: int, out_channel: int, kernel_size: int = 3, stride=1, groups: int = 1,
                 use_norm: Optional[bool] = False, bias=False, use_act: Optional[bool] = False,
                 dilation: int = 1, norm_type='BN'):
        super(ConvNormAct, self).__init__()

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(stride, int):
            stride = (stride, stride)
        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        padding = (int((kernel_size[0] - 1) / 2) * dilation[0], int((kernel_size[1] - 1) / 2) * dilation[1])

        if use_norm:
            bias = False

        self.conv = torch.nn.Conv2d(in_channel, out_channel, kernel_size, stride, padding,
                                    dilation=dilation, groups=groups, bias=bias)

        # self.norm_layer = nn.BatchNorm2d(out_channel, momentum=momentum) if use_norm else None

        self.norm_layer = nn.BatchNorm2d(out_channel) if use_norm else None
        # else:
        #     self.norm_layer = LayerNorm(out_channel, data_format="channels_first") if use_norm else None

        self.act = nn.SiLU(inplace=True) if use_act else None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        if self.norm_layer is not None:
            x = self.norm_layer(x)
        if self.act is not None:
            x = self.act(x)
        return x


class InvertedResidual(nn.Module):
    """
    Inverted residual block (MobileNetv2): https://arxiv.org/abs/1801.04381
    """

    def __init__(self,
                 in_channel: int,
                 out_channel: int,
                 stride: int,
                 expand_ratio: Union[int, float] = 4,
                 kernel_size=3,
                 dilation: int = 1,
                 use_se=False,
                 out_size: tuple = (224, 224),
                 norm_type="BN"
                 ) -> None:
        assert stride in [1, 2]
        super(InvertedResidual, self).__init__()
        self.stride = stride
        int_size = (int(out_size[0] * stride), int(out_size[1] * stride))

        expansion_channel = int(round(in_channel * expand_ratio))
        self.use_res_connect = self.stride == 1 and in_channel == out_channel

        block = nn.Sequential()
        if expand_ratio != 1:
            block.add_module(name="exp_1x1",
                             module=ConvNormAct(in_channel=in_channel, out_channel=expansion_channel, kernel_size=1,
                                                use_act=True, use_norm=True, norm_type=norm_type))
        block.add_module(
            name="conv_3x3",
            module=ConvNormAct(in_channel=expansion_channel, out_channel=expansion_channel, stride=stride,
                               kernel_size=kernel_size, groups=expansion_channel, use_act=True, use_norm=True,
                               dilation=dilation, norm_type=norm_type))

        block.add_module(name="red_1x1",
                         module=ConvNormAct(in_channel=expansion_channel, out_channel=out_channel, kernel_size=1,
                                            use_act=False, use_norm=True, norm_type=norm_type))

        self.block = block

    def forward(self, x):
        if self.use_res_connect:
            return x + self.block(x)
        else:
            return self.block(x)


@HEADS.register_module()
class LcTFCNHead(BaseDecodeHead):
    """Fully Convolution Networks for Semantic Segmentation.

    This head is implemented of `FCNNet <https://arxiv.org/abs/1411.4038>`_.

    Args:
        num_convs (int): Number of convs in the head. Default: 2.
        kernel_size (int): The kernel size for convs in the head. Default: 3.
        concat_input (bool): Whether concat the input and output of convs
            before classification layer.
        dilation (int): The dilation rate for convs in the head. Default: 1.
    """

    def __init__(self,
                 num_convs=2,
                 kernel_size=3,
                 fcn_ratio=2,
                 concat_input=True,
                 dilation=1,
                 **kwargs):
        assert num_convs >= 0 and dilation > 0 and isinstance(dilation, int)
        self.num_convs = num_convs
        self.concat_input = concat_input
        self.kernel_size = kernel_size
        super(LcTFCNHead, self).__init__(**kwargs)
        if num_convs == 0:
            assert self.in_channels == self.channels

        conv_padding = (kernel_size // 2) * dilation
        convs = []
        # convs.append(
        #     ConvModule(
        #         self.in_channels,
        #         self.channels,
        #         kernel_size=kernel_size,
        #         padding=conv_padding,
        #         dilation=dilation,
        #         conv_cfg=self.conv_cfg,
        #         norm_cfg=self.norm_cfg,
        #         act_cfg=self.act_cfg))
        convs.append(
            InvertedResidual(in_channel=self.in_channels, out_channel=self.channels, expand_ratio=fcn_ratio, stride=1))
        for i in range(num_convs - 1):
            convs.append(
                ConvModule(
                    self.channels,
                    self.channels,
                    kernel_size=kernel_size,
                    padding=conv_padding,
                    dilation=dilation,
                    conv_cfg=self.conv_cfg,
                    norm_cfg=self.norm_cfg,
                    act_cfg=self.act_cfg))
        if num_convs == 0:
            self.convs = nn.Identity()
        else:
            self.convs = nn.Sequential(*convs)
        if self.concat_input:
            self.conv_cat = ConvModule(
                self.in_channels + self.channels,
                self.channels,
                kernel_size=kernel_size,
                padding=kernel_size // 2,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg)

    def forward(self, inputs):
        """Forward function."""
        x = self._transform_inputs(inputs)
        output = self.convs(x)
        if self.concat_input:
            output = self.conv_cat(torch.cat([x, output], dim=1))
        output = self.cls_seg(output)
        return output
