"""
LcT: Locally-enhanced Cross-window Vision Transformer
"""

import torch.nn as nn
import torch

from typing import Optional, Union, Tuple, Dict, Callable
from einops import rearrange
from torch import Tensor
from torch.nn import functional as F
from timm.models.layers import DropPath, to_2tuple, trunc_normal_


class LayerNorm(nn.Module):
    r""" LayerNorm that supports two data formats: channels_last (default) or channels_first.
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs
    with shape (batch_size, channels, height, width).

    this code comes from convnext
    """

    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x):
        if self.data_format == "channels_last":
            return F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None] * x + self.bias[:, None, None]
            return x


class ConvNormAct(nn.Module):
    def __init__(self,
                 in_channel: int, out_channel: int, kernel_size: int = 3, stride=1, groups: int = 1,
                 use_norm: Optional[bool] = False, bias=False, use_act: Optional[bool] = False,
                 dilation: int = 1, norm_type='BN'):
        super(ConvNormAct, self).__init__()

        if isinstance(kernel_size, int):
            kernel_size = (kernel_size, kernel_size)
        if isinstance(stride, int):
            stride = (stride, stride)
        if isinstance(dilation, int):
            dilation = (dilation, dilation)

        padding = (int((kernel_size[0] - 1) / 2) * dilation[0], int((kernel_size[1] - 1) / 2) * dilation[1])

        if use_norm:
            bias = False

        self.conv = torch.nn.Conv2d(in_channel, out_channel, kernel_size, stride, padding,
                                    dilation=dilation, groups=groups, bias=bias)

        assert norm_type in ['BN', 'LN'], "norm_type must be 'LN'(layer norm) of 'BN'(batch norm)"
        if norm_type == 'BN':
            self.norm_layer = nn.BatchNorm2d(out_channel) if use_norm else None
        elif norm_type == 'LN':
            self.norm_layer = LayerNorm(out_channel, data_format="channels_first") if use_norm else None

        self.act = nn.SiLU(inplace=True) if use_act else None

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv(x)
        if self.norm_layer is not None:
            x = self.norm_layer(x)
        if self.act is not None:
            x = self.act(x)
        return x


class SqueezeExcitation(nn.Module):
    def __init__(self, input_channels: int, squeeze_channels: int):
        super().__init__()
        self.fc1 = nn.Conv2d(input_channels, squeeze_channels, 1)
        self.fc2 = nn.Conv2d(squeeze_channels, input_channels, 1)

    def _scale(self, input: Tensor) -> Tensor:
        scale = F.adaptive_avg_pool2d(input, 1)
        scale = self.fc1(scale)
        scale = F.silu(scale, inplace=True)
        scale = self.fc2(scale)
        return scale.sigmoid()

    def forward(self, input: Tensor) -> Tensor:
        scale = self._scale(input)
        return scale * input


class InvertedResidual(nn.Module):
    """
    Inverted residual block (MobileNetv2): https://arxiv.org/abs/1801.04381
    """

    def __init__(self,
                 in_channel: int,
                 out_channel: int,
                 stride: int,
                 expand_ratio: Union[int, float] = 4,
                 kernel_size=3,
                 dilation: int = 1,
                 use_se=False,

                 norm_type="BN"
                 ) -> None:
        assert stride in [1, 2]
        super(InvertedResidual, self).__init__()
        self.stride = stride

        expansion_channel = int(round(in_channel * expand_ratio))
        self.use_res_connect = self.stride == 1 and in_channel == out_channel

        block = nn.Sequential()
        if expand_ratio != 1:
            block.add_module(name="exp_1x1",
                             module=ConvNormAct(in_channel=in_channel, out_channel=expansion_channel, kernel_size=1,
                                                use_act=True, use_norm=True, norm_type=norm_type))
        block.add_module(
            name="conv_3x3",
            module=ConvNormAct(in_channel=expansion_channel, out_channel=expansion_channel, stride=stride,
                               kernel_size=kernel_size, groups=expansion_channel, use_act=True, use_norm=True,
                               dilation=dilation, norm_type=norm_type))

        if use_se:
            # squeeze and excitation
            squeeze_channels = max(1, in_channel // 4)
            block.add_module(name="se",
                             module=SqueezeExcitation(expansion_channel, squeeze_channels))

        block.add_module(name="red_1x1",
                         module=ConvNormAct(in_channel=expansion_channel, out_channel=out_channel, kernel_size=1,
                                            use_act=False, use_norm=True, norm_type=norm_type))

        self.block = block

    def forward(self, x: Tensor) -> Tensor:
        if self.use_res_connect:
            return x + self.block(x)
        else:
            return self.block(x)


class MultiHeadAttention(nn.Module):
    """
            This layer applies a multi-head attention as described in "Attention is all you need" paper
            https://arxiv.org/abs/1706.03762
    """

    def __init__(self, embed_dim: int, num_heads: int, attn_dropout: Optional[float] = 0.0,
                 bias: Optional[bool] = True, window_size=2,
                 *args, **kwargs):
        """
        :param embed_dim: Embedding dimension
        :param num_heads: Number of attention heads
        :param attn_dropout: Attention dropout
        :param bias: Bias
        """
        super(MultiHeadAttention, self).__init__()

        assert embed_dim % num_heads == 0, "Got: embed_dim={} and num_heads={}".format(embed_dim, num_heads)

        # self.attn_dropout = nn.Dropout(p=attn_dropout, inplace=True)

        self.to_qkv = nn.Conv2d(embed_dim, embed_dim * 3, 1, bias=False)  # shuffle transformer bias=False
        self.out_proj = nn.Conv2d(embed_dim, embed_dim, 1, bias=True)

        self.head_dim = embed_dim // num_heads
        self.scaling = self.head_dim ** -0.5
        self.softmax = nn.Softmax(dim=-1)
        self.num_heads = num_heads
        self.patch_size = window_size
        self.embed_dim = embed_dim

        print('num head', num_heads)

    def forward(self, x: Tensor) -> Tensor:
        B, C, H, W = x.shape
        x = self.to_qkv(x)
        # B, 3C, H, W -->[3 B*num h  N  d]
        query, key, value = rearrange(x, 'b (qkv h d) (hh ws1) (ww ws2) -> qkv (b ws1 ws2) h (hh ww) d',
                                      ws1=self.patch_size, ws2=self.patch_size, qkv=3, h=self.num_heads)
        query = query * self.scaling

        # QK^T
        # [B*num, h, N, d] x [B*num h d N] --> [B*num h N N]
        attn = torch.matmul(query, key.transpose(2, 3))

        attn = self.softmax(attn)
        # attn = self.attn_dropout(attn)

        # attn @ v
        # [B*num h N N] x [B*num h N d] --> [B*num h N d]
        out = torch.matmul(attn, value)

        # [B*num, h, N, d] -->B, C, H, W
        out = rearrange(out, '(b ws1 ws2) h (hh ww) d -> b (h d) (hh ws1) (ww ws2)',
                        ws1=self.patch_size, ws2=self.patch_size, h=self.num_heads, b=B, hh=H // self.patch_size)

        out = self.out_proj(out)

        return out


class TransformerEncoder(nn.Module):
    """
        This class defines the Transformer encoder (pre-norm) as described in "Attention is all you need" paper
            https://arxiv.org/abs/1706.03762
    """

    def __init__(self, embed_dim: int, num_heads: Optional[int] = 8, window_size=2, kernel_size=3,
                 attn_dropout=0.0, dropout=0.0, drop_path=0.1, ffn_ratio=2,
                 use_rpe=False, ffn_type='',
                 ):
        super(TransformerEncoder, self).__init__()

        print("drop_path", drop_path)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

        self.mhsa = nn.Sequential(nn.BatchNorm2d(embed_dim),
                                  # LayerNorm(normalized_shape=embed_dim, data_format='channels_first'),
                                  MultiHeadAttention(embed_dim, num_heads, attn_dropout=attn_dropout, bias=True,
                                                     window_size=window_size),
                                  # nn.Dropout(p=dropout, inplace=True)
                                  )
        if ffn_type == 'inverted_residual':
            self.ffn = nn.Sequential(nn.BatchNorm2d(embed_dim),
                                     # LayerNorm(normalized_shape=embed_dim, data_format='channels_first'),
                                     InvertedResidual(in_channel=embed_dim, out_channel=embed_dim,
                                                      stride=1, expand_ratio=ffn_ratio, kernel_size=3, ),
                                     # nn.Dropout(p=dropout, inplace=True)
                                     )
        elif ffn_type == 'mlp':
            self.ffn = nn.Sequential(nn.BatchNorm2d(embed_dim),
                                     nn.Conv2d(in_channels=embed_dim, out_channels=int(ffn_ratio * embed_dim),
                                               kernel_size=1, stride=1, padding=0),
                                     nn.SiLU(),
                                     nn.Conv2d(in_channels=int(2 * embed_dim), out_channels=embed_dim,
                                               kernel_size=1, stride=1, padding=0))

        else:
            raise Exception("error ffn type, only support 'inverted_residual' or 'mlp' ")

    def forward(self, x: Tensor) -> Tensor:
        # Multi-head attention
        x = x + self.drop_path(self.mhsa(x))

        # Feed forward network
        x = x + self.drop_path(self.ffn(x))
        return x


class LcTBlock(nn.Module):
    """

    """

    def __init__(self, drop_paths, transformer_dim: int, ffn_ratio: int,
                 transformer_blocks: Optional[int] = 2,
                 head_dim: Optional[int] = 32,
                 drop_rate: Optional[float] = 0.0, attn_dropout: Optional[float] = 0.0,
                 window_h: Optional[int] = 8, window_w: Optional[int] = 8,
                 conv_ksize: Optional[int] = 3, ffn_ksize=3,
                 dilation: Optional[int] = 1, no_fusion: Optional[bool] = False, ffn_type='inverted_residual',
                 down_sample=None,
                 *args, **kwargs):

        super(LcTBlock, self).__init__()
        assert transformer_dim % head_dim == 0
        num_heads = transformer_dim // head_dim

        self.local_rep = ConvNormAct(in_channel=transformer_dim, out_channel=transformer_dim,
                                     kernel_size=conv_ksize, stride=1, use_norm=True, use_act=True, dilation=dilation,
                                     norm_type='LN')

        global_rep = [TransformerEncoder(embed_dim=transformer_dim, num_heads=num_heads, window_size=window_w,
                                         kernel_size=ffn_ksize,
                                         attn_dropout=attn_dropout, dropout=drop_rate, drop_path=drop_paths[i],
                                         ffn_type=ffn_type, ffn_ratio=ffn_ratio)
                      for i in range(transformer_blocks)]
        self.global_rep = nn.Sequential(*global_rep)

        self.fusion = InvertedResidual(in_channel=2 * transformer_dim, out_channel=transformer_dim,
                                       kernel_size=conv_ksize, stride=1, expand_ratio=1)

        self.down_sample = down_sample

    def forward(self, x: Tensor) -> Tensor:
        res = x

        x = self.local_rep(x)

        # learn global representations
        x = self.global_rep(x)

        if self.fusion is not None:
            x = self.fusion(
                torch.cat((res, x), dim=1))

        if self.down_sample is not None:
            x = self.down_sample(x)

        return x


class LcT(nn.Module):
    """
        LcT: Locally-enhanced Cross-window Vision Transformer
    """

    def __init__(self, num_classes=1000, drop_rate=0.0, drop_path_rate=0.1, classifier_drop_rate=0.1, attn_dropout=0.0,
                 ffn_type='inverted_residual', stem_type='conv', heads_dim=24,
                 ffn_ratio=2, irb_ratio=4, output_ratio=4,
                 window_size=[7, 4, 2, 1], kernel_size=[3, 3, 3, 3], ffn_kernel_size=[3, 3, 3, 3], depths=[2, 2, 6, 2],
                 stem_channels=[16, 32, 64, 96], transformer_dims=[96, 144, 192, 240],
                 qkv_bias=True, qk_scale=None,
                 *args, **kwargs) -> None:
        super(LcT, self).__init__()

        # stochastic depth
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule

        # stem
        self.stem = self.stem_block(stem_type=stem_type, stem_channels=stem_channels, ratio=irb_ratio)

        # stages
        self.stages = nn.ModuleList()
        for i in range(len(depths)):
            down_sample = None
            if i != len(depths) - 1:
                down_sample = InvertedResidual(in_channel=transformer_dims[i], out_channel=transformer_dims[i + 1],
                                               stride=2, expand_ratio=irb_ratio, use_se=False)

            stage = LcTBlock(drop_paths=dpr[sum(depths[:i]):sum(depths[:i + 1])],
                             transformer_dim=transformer_dims[i],
                             ffn_ratio=ffn_ratio,
                             transformer_blocks=depths[i],
                             head_dim=heads_dim,
                             drop_rate=drop_rate,
                             attn_dropout=attn_dropout,
                             window__h=window_size[i],
                             window_w=window_size[i],
                             conv_ksize=kernel_size[i],
                             ffn_ksize=ffn_kernel_size[i],
                             ffn_type=ffn_type,
                             down_sample=down_sample,
                             )

            self.stages.append(stage)

        # last conv 1x1 out
        out_channel = transformer_dims[-1] * output_ratio
        self.last_conv_1x1 = ConvNormAct(in_channel=transformer_dims[-1], out_channel=out_channel,
                                         kernel_size=1, use_act=True, use_norm=True, )

        # classifier
        self.classifier = nn.Sequential()
        self.classifier.add_module(name="global_pool", module=nn.AvgPool2d(kernel_size=7))
        if 0.0 < classifier_drop_rate < 1.0:
            self.classifier.add_module(name="dropout", module=nn.Dropout(p=classifier_drop_rate, inplace=True))
        self.classifier.add_module(name="fc",
                                   module=nn.Linear(in_features=out_channel, out_features=num_classes, bias=True)
                                   )

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight, mode='fan_out')
            if m.bias is not None:
                nn.init.zeros_(m.bias)
        elif isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.02)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, (nn.LayerNorm, nn.BatchNorm2d)):
            nn.init.constant_(m.bias, 0)
            nn.init.constant_(m.weight, 1.0)

    @staticmethod
    def stem_block(stem_type, stem_channels, ratio):

        channels = stem_channels

        if stem_type == "inverted_residual_stem":
            block = [ConvNormAct(in_channel=3, out_channel=16, stride=2, kernel_size=3, use_norm=True, use_act=True),
                     InvertedResidual(in_channel=channels[0], out_channel=channels[1], stride=1, expand_ratio=ratio),
                     InvertedResidual(in_channel=channels[1], out_channel=channels[2], stride=2, expand_ratio=ratio),
                     InvertedResidual(in_channel=channels[2], out_channel=channels[2], stride=1, expand_ratio=ratio),
                     InvertedResidual(in_channel=channels[2], out_channel=channels[2], stride=1, expand_ratio=ratio), ]

        elif stem_type == "conv":

            block = [
                ConvNormAct(in_channel=3, out_channel=channels[0], stride=2, kernel_size=3, use_norm=True, use_act=True,
                            ),
                ConvNormAct(in_channel=channels[0], out_channel=channels[1], stride=1, kernel_size=3, use_norm=True,
                            use_act=True,
                            ),
                ConvNormAct(in_channel=channels[1], out_channel=channels[2], stride=1, kernel_size=3, use_norm=True,
                            use_act=True,
                            ),
                ConvNormAct(in_channel=channels[2], out_channel=channels[3], stride=2, kernel_size=3, use_norm=True,
                            use_act=False),
            ]

        return nn.Sequential(*block)

    def extract_features(self, x: Tensor) -> Tensor:

        x = self.stem(x)
        for stage in self.stages:
            x = stage(x)

        return x

    def forward(self, x: Tensor) -> Tensor:
        x = self.extract_features(x)
        x = self.last_conv_1x1(x)
        for i, blk in enumerate(self.classifier):
            if i == 0:
                x = blk(x).flatten(1)
            else:
                x = blk(x)
        return x


def load_pretrained(model, path):
    checkpoint = torch.load(path, map_location='cpu')
    msg = model.load_state_dict(checkpoint['model'], strict=False)
    print(msg)


def LcT_Tiny(pretrained=False, checkpoint_path=None, **kwargs):
    model = LcT(heads_dim=12, irb_ratio=2, output_ratio=3,
                depths=[2, 2, 4, 2],
                stem_channels=[16, 24, 36, 36], transformer_dims=[36, 72, 108, 144], **kwargs)
    if pretrained:
        if checkpoint_path is None:
            path = "./weight/LcT_Tiny.pth"
        else:
            path = checkpoint_path
        load_pretrained(model, path)
    return model


def LcT_Small(pretrained=False, checkpoint_path=None, **kwargs):
    model = LcT(heads_dim=24, irb_ratio=3, output_ratio=3,
                depths=[2, 2, 4, 2],
                stem_channels=[16, 32, 48, 48], transformer_dims=[48, 96, 144, 192], **kwargs)
    if pretrained:
        if checkpoint_path is None:
            path = "./weight/LcT_Small.pth"
        else:
            path = checkpoint_path
        load_pretrained(model, path)
    return model


def LcT_Base(pretrained=False, checkpoint_path=None, **kwargs):
    model = LcT(heads_dim=24, irb_ratio=4, output_ratio=4,
                depths=[2, 2, 6, 2],
                stem_channels=[16, 32, 64, 96], transformer_dims=[96, 144, 192, 240], **kwargs)
    if pretrained:
        if checkpoint_path is None:
            path = "./weight/LcT_Base.pth"
        else:
            path = checkpoint_path
        load_pretrained(model, path)
    return model
