# --------------------------------------------------------
# These codes are based on Swin Transformer

# --------------------------------------------------------
import multiprocessing
import os
import time
import argparse
import datetime
import torch

from timm.utils import accuracy, AverageMeter, ModelEma
from config import get_config
from data.build import build_dataset
from model.build_model import build_model
from logger import create_logger
import numpy as np
from data.samplers import SubsetRandomSampler


def parse_option():
    parser = argparse.ArgumentParser('LcT training and evaluation script', add_help=False)

    parser.add_argument('--cfg', type=str, default="./configs/LcT_Small.yaml", help='path to config file')
    parser.add_argument('--training_type', type=str, default="simple", help="simple or other")

    parser.add_argument('--image-size', type=int, default=224, help="image size")
    parser.add_argument('--batch-size', type=int, default=128, help="batch size for single GPU")
    parser.add_argument('--data-path', type=str, default='./data/imagenet', help='path to dataset')

    parser.add_argument("--opts", help="Modify config options by adding 'KEY VALUE' pairs. ", default=None, nargs='+', )
    parser.add_argument('--zip', action='store_true', help='use zipped dataset instead of folder dataset')
    parser.add_argument('--cache-mode', type=str, default='part', choices=['no', 'full', 'part'],
                        help='no: no cache, '
                             'full: cache all data, '
                             'part: sharding the dataset into nonoverlapping pieces and only cache one piece')
    parser.add_argument('--checkpoint', default="./weight/LcT_Small.pth", help='checkpoint path')
    parser.add_argument('--resume', help='resume from checkpoint')
    parser.add_argument('--accumulation-steps', type=int, help="gradient accumulation steps")
    parser.add_argument('--use-checkpoint', action='store_true',
                        help="whether to use gradient checkpointing to save memory")
    parser.add_argument('--amp-opt-level', type=str, default='O0', choices=['O0', 'O1', 'O2'],
                        help='mixed precision opt level, if O0, no amp is used')
    parser.add_argument('--output', default='output', type=str, metavar='PATH',
                        help='root of output folder, the full path is <output>/<model_name>/<tag> (default: output)')
    parser.add_argument('--tag', help='tag of experiment')
    parser.add_argument('--eval', action='store_true', help='Perform evaluation only')
    parser.add_argument('--throughput', action='store_true', help='Test throughput only')

    parser.add_argument('--use_ddp', type=bool, default=False)
    parser.add_argument("--local_rank", type=int, default=0, required=False, help='local rank for DDP')
    parser.add_argument("--gpu_id", type=int, default=0)
    parser.add_argument("--num_gpus", type=int, default=1)
    parser.add_argument("--world_size", type=int, default=-1)

    # training_setting
    parser.add_argument("--num_workers", type=int, default=8)
    parser.add_argument("--random_seed", type=int, default=0)

    # --------------------------------------------------------
    temp_args, _ = parser.parse_known_args()
    config = get_config(temp_args)

    opts = parser.parse_args()

    return config, opts


def main(opts, config, **kwargs):
    # --------------------------------------------------------
    # --------------------------------------------------------
    os.makedirs(config.OUTPUT, exist_ok=True)
    logger = create_logger(output_dir=config.OUTPUT, dist_rank=0, name=f"{config.MODEL.NAME}")

    path = os.path.join(config.OUTPUT, "config.json")
    with open(path, "w") as f:
        f.write(config.dump())
    logger.info(f"Full config saved to {path}")
    # print config
    # logger.info(config.dump())

    torch.cuda.set_device(opts.gpu_id)
    # --------------------------------------------------------
    # --------------------------------------------------------
    # load dataset

    dataset_val, _ = build_dataset(is_train=False, config=config)
    indices = np.arange(0, len(dataset_val))
    sampler_val = SubsetRandomSampler(indices)
    data_loader_val = torch.utils.data.DataLoader(
        dataset_val, sampler=sampler_val,
        batch_size=opts.batch_size,
        shuffle=False,
        num_workers=config.DATA.NUM_WORKERS,
        pin_memory=config.DATA.PIN_MEMORY,
        drop_last=False
    )

    # create model
    model = build_model(opts, config)
    model.cuda()  # model.to(rank)
    logger.info(str(model))

    # create optimizer
    model_without_ddp = model

    n_parameters = sum(p.numel() for p in model.parameters() if p.requires_grad)
    logger.info(f"number of params: {n_parameters / 1e6} M")
    if hasattr(model_without_ddp, 'flops'):
        flops = model_without_ddp.flops()
        logger.info(f"number of GFLOPs: {flops / 1e9} G")

    if opts.checkpoint:
        logger.info(f'auto resuming from {opts.checkpoint}')
        checkpoint = torch.load(opts.checkpoint, map_location='cpu')
        msg = model.load_state_dict(checkpoint['model'], strict=False)
        logger.info(msg)
        acc1, acc5, loss = validate(config, data_loader_val, model, logger, 0)
        logger.info(f"Accuracy of the network on the {len(dataset_val)} test images: {acc1:.1f}%")
    else:
        logger.info("Checkpoint is None, only test throughput")

    throughput(data_loader_val, model, logger)
    return


@torch.no_grad()
def validate(config, data_loader, model, logger, rank, ema_mode=False):
    criterion = torch.nn.CrossEntropyLoss()
    model.eval()

    batch_time = AverageMeter()
    loss_meter = AverageMeter()
    acc1_meter = AverageMeter()
    acc5_meter = AverageMeter()

    end = time.time()
    for idx, (images, target) in enumerate(data_loader):
        images = images.cuda(non_blocking=True)  # images = images.to(rank)
        target = target.cuda(non_blocking=True)  # target = target.to(rank)

        # compute output
        output = model(images)

        # measure accuracy and record loss
        loss = criterion(output, target)
        acc1, acc5 = accuracy(output, target, topk=(1, 5))

        # acc1 = reduce_tensor(acc1)
        # acc5 = reduce_tensor(acc5)
        # loss = reduce_tensor(loss)

        loss_meter.update(loss.item(), target.size(0))
        acc1_meter.update(acc1.item(), target.size(0))
        acc5_meter.update(acc5.item(), target.size(0))

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if idx % config.PRINT_FREQ == 0:
            memory_used = torch.cuda.max_memory_allocated() / (1024.0 * 1024.0)
            logger.info(
                f'Test: [{idx}/{len(data_loader)}]\t'
                f'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                f'Loss {loss_meter.val:.4f} ({loss_meter.avg:.4f})\t'
                f'Acc@1 {acc1_meter.val:.3f} ({acc1_meter.avg:.3f})\t'
                f'Acc@5 {acc5_meter.val:.3f} ({acc5_meter.avg:.3f})\t'
                f'Mem {memory_used:.0f}MB')

    if ema_mode:
        logger.info(f' EMA Acc@1 {acc1_meter.avg:.3f} Acc@5 {acc5_meter.avg:.3f}')
    else:
        logger.info(f' * Acc@1 {acc1_meter.avg:.3f} Acc@5 {acc5_meter.avg:.3f}')

    return acc1_meter.avg, acc5_meter.avg, loss_meter.avg


@torch.no_grad()
def throughput(data_loader, model, logger):
    model.eval()

    for idx, (images, _) in enumerate(data_loader):
        images = images.cuda(non_blocking=True)
        batch_size = images.shape[0]
        for i in range(50):
            model(images)
        torch.cuda.synchronize()
        logger.info(f"throughput averaged with 30 times")
        tic1 = time.time()
        for i in range(30):
            model(images)
        torch.cuda.synchronize()
        tic2 = time.time()
        logger.info(f"batch_size {batch_size} throughput {30 * batch_size / (tic2 - tic1)}")
        return


def main_worker(**kwargs):
    config, opts = parse_option()
    n_cpus = multiprocessing.cpu_count()
    setattr(opts, "num_workers", n_cpus)
    setattr(opts, "device_id", None)
    main(opts=opts, config=config, **kwargs)


if __name__ == "__main__":
    main_worker()
