# --------------------------------------------------------
# These codes are based on Swin Transformer
# --------------------------------------------------------
import multiprocessing
import os
import time
import argparse
import datetime
import torch

import torch.distributed as dist
from timm.loss import LabelSmoothingCrossEntropy, SoftTargetCrossEntropy
from timm.utils import accuracy, AverageMeter, ModelEma
from torch.cuda import amp
from config import get_config
from model.build_model import build_model
from data import build_loader
from lr_scheduler import build_scheduler
from optimizer import build_optimizer
from logger import create_logger
from utils import load_checkpoint, save_checkpoint, get_grad_norm, auto_resume_helper, reduce_tensor, device_setup, \
    distributed_init


def parse_option():
    parser = argparse.ArgumentParser('LcT training and evaluation script', add_help=False)

    parser.add_argument('--cfg', type=str, default="./configs/LcT_Small.yaml", help='path to config file')
    parser.add_argument('--training_type', type=str, default="simple", help="simple or other")

    parser.add_argument('--image-size', type=int, default=224, help="image size")
    parser.add_argument('--batch-size', type=int, default=350, help="batch size for single GPU")
    parser.add_argument('--data-path', type=str, default='./data/imagenet', help='path to dataset')

    parser.add_argument("--opts", help="Modify config options by adding 'KEY VALUE' pairs. ", default=None, nargs='+', )
    parser.add_argument('--zip', action='store_true', help='use zipped dataset instead of folder dataset')
    parser.add_argument('--cache-mode', type=str, default='part', choices=['no', 'full', 'part'],
                        help='no: no cache, '
                             'full: cache all data, '
                             'part: sharding the dataset into nonoverlapping pieces and only cache one piece')
    parser.add_argument('--resume', default="./weight/LcT_Small.pth", help='resume from checkpoint')
    parser.add_argument('--accumulation-steps', type=int, help="gradient accumulation steps")
    parser.add_argument('--use-checkpoint', action='store_true',
                        help="whether to use gradient checkpointing to save memory")
    parser.add_argument('--amp-opt-level', type=str, default='O0', choices=['O0', 'O1', 'O2'],
                        help='mixed precision opt level, if O0, no amp is used')
    parser.add_argument('--output', default='output', type=str, metavar='PATH',
                        help='root of output folder, the full path is <output>/<model_name>/<tag> (default: output)')
    parser.add_argument('--tag', help='tag of experiment')
    parser.add_argument('--eval', action='store_true', help='Perform evaluation only')
    parser.add_argument('--throughput', action='store_true', help='Test throughput only')

    # distributed training
    parser.add_argument('--use_ddp', type=bool, default=False)
    parser.add_argument("--local_rank", type=int, default=0, required=False, help='local rank for DDP')
    parser.add_argument("--num_gpus", type=int, default=1)
    parser.add_argument("--world_size", type=int, default=-1)
    #  parser.add_argument("--dist_port", type=int, default=30786)

    # training_setting
    parser.add_argument("--num_workers", type=int, default=8)
    parser.add_argument("--random_seed", type=int, default=0)

    # --------------------------------------------------------
    temp_args, _ = parser.parse_known_args()
    config = get_config(temp_args)

    opts = parser.parse_args()

    return config, opts


def main(opts, config, **kwargs):

    rank = dist.get_rank()

    linear_scaled_lr = config.TRAIN.BASE_LR * config.DATA.BATCH_SIZE * dist.get_world_size() / 512.0
    linear_scaled_warmup_lr = config.TRAIN.WARMUP_LR * config.DATA.BATCH_SIZE * dist.get_world_size() / 512.0
    linear_scaled_min_lr = config.TRAIN.MIN_LR * config.DATA.BATCH_SIZE * dist.get_world_size() / 512.0
    # gradient accumulation also need to scale the learning rate
    if config.TRAIN.ACCUMULATION_STEPS > 1:
        linear_scaled_lr = linear_scaled_lr * config.TRAIN.ACCUMULATION_STEPS
        linear_scaled_warmup_lr = linear_scaled_warmup_lr * config.TRAIN.ACCUMULATION_STEPS
        linear_scaled_min_lr = linear_scaled_min_lr * config.TRAIN.ACCUMULATION_STEPS
    config.defrost()
    config.TRAIN.BASE_LR = linear_scaled_lr
    config.TRAIN.WARMUP_LR = linear_scaled_warmup_lr
    config.TRAIN.MIN_LR = linear_scaled_min_lr
    config.freeze()
    # --------------------------------------------------------
    # --------------------------------------------------------
    os.makedirs(config.OUTPUT, exist_ok=True)
    logger = create_logger(output_dir=config.OUTPUT, dist_rank=dist.get_rank(), name=f"{config.MODEL.NAME}")

    path = os.path.join(config.OUTPUT, "config.json")
    with open(path, "w") as f:
        f.write(config.dump())
    logger.info(f"Full config saved to {path}")
    # print config
    logger.info(config.dump())

    # --------------------------------------------------------
    # --------------------------------------------------------
    # load dataset
    dataset_train, dataset_val, data_loader_train, data_loader_val, mixup_fn = build_loader(config)

    # create model
    model = build_model(opts, config)
    model.cuda()  # model.to(rank)
    logger.info(str(model))

    # EMA
    model_ema = None
    if config.EMA.USE:
        # Important to create EMA model after cuda(), DP wrapper, and AMP but before SyncBN and DDP wrapper
        model_ema = ModelEma(
            model,
            decay=config.EMA.DECAY_RATE,
            device='',
            resume='')
        print("Using EMA. decay rate = %.6f" % config.EMA.DECAY_RATE)

    # create optimizer
    optimizer = build_optimizer(config, model)

    model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[rank])
    model_without_ddp = model.module

    n_parameters = sum(p.numel() for p in model.parameters() if p.requires_grad)
    logger.info(f"number of params: {n_parameters / 1e6} M")
    if hasattr(model_without_ddp, 'flops'):
        flops = model_without_ddp.flops()
        logger.info(f"number of GFLOPs: {flops / 1e9} G")

    lr_scheduler = build_scheduler(config, optimizer, len(data_loader_train))

    if config.AUG.MIXUP > 0.:
        # smoothing is handled with mixup label transform
        criterion = SoftTargetCrossEntropy()
    elif config.MODEL.LABEL_SMOOTHING > 0.:
        criterion = LabelSmoothingCrossEntropy(smoothing=config.MODEL.LABEL_SMOOTHING)
    else:
        criterion = torch.nn.CrossEntropyLoss()

    max_accuracy = 0.0

    if config.TRAIN.AUTO_RESUME:
        resume_file = auto_resume_helper(config.OUTPUT)
        if resume_file:
            if config.MODEL.RESUME:
                logger.warning(f"auto-resume changing resume file from {config.MODEL.RESUME} to {resume_file}")
            config.defrost()
            config.MODEL.RESUME = resume_file
            config.freeze()
            logger.info(f'auto resuming from {resume_file}')
        else:
            logger.info(f'no checkpoint found in {config.OUTPUT}, ignoring auto resume')

    if config.MODEL.RESUME:
        max_accuracy = load_checkpoint(config, model_without_ddp, optimizer, lr_scheduler, logger, model_ema)
        acc1, acc5, loss = validate(config, data_loader_val, model, logger, rank)
        logger.info(f"Accuracy of the network on the {len(dataset_val)} test images: {acc1:.1f}%")

        if config.EMA.USE:
            validate_ema(config, data_loader_val, model_ema.ema, logger, rank, ema_mode=True)
        #     # logger.info(f"Accuracy of the EMA model on the {len(dataset_val)} test images: {acc1:.1f}%")

        if config.EVAL_MODE:
            return

    if config.THROUGHPUT_MODE:
        throughput(data_loader_val, model, logger)
        return
    # --------------------------------------------------------
    # train
    # --------------------------------------------------------
    logger.info("Start training")
    start_time = time.time()
    start_epoch = config.TRAIN.START_EPOCH

    for epoch in range(config.TRAIN.START_EPOCH, config.TRAIN.EPOCHS):

        data_loader_train.sampler.set_epoch(epoch)

        train_one_epoch(config, model, criterion, data_loader_train, optimizer, epoch, mixup_fn, lr_scheduler, logger,
                        rank, model_ema)
        if dist.get_rank() == 0 and (epoch % config.SAVE_FREQ == 0 or epoch == (config.TRAIN.EPOCHS - 1)):
            # if epoch % config.SAVE_FREQ == 0 or epoch == (config.TRAIN.EPOCHS - 1):
            save_checkpoint(config, epoch, model_without_ddp, max_accuracy, optimizer, lr_scheduler, logger, model_ema)

        acc1, acc5, loss = validate(config, data_loader_val, model, logger, rank)
        logger.info(f"Accuracy of the network on the {len(dataset_val)} test images: {acc1:.1f}%")
        max_accuracy = max(max_accuracy, acc1)
        logger.info(f'Max accuracy: {max_accuracy:.2f}%')

        # test on model_ema
        if config.EMA.USE and epoch % 5 == 0:
            validate_ema(config, data_loader_val, model_ema.ema, logger, rank, ema_mode=True)

    total_time = time.time() - start_time
    total_time_str = str(datetime.timedelta(seconds=int(total_time)))
    logger.info('Training time {}'.format(total_time_str))


def train_one_epoch(config, model, criterion, data_loader, optimizer, epoch, mixup_fn, lr_scheduler, logger, rank,
                    model_ema=None, ):
    model.train()
    optimizer.zero_grad()

    # 混合精度
    enable_amp = True
    scaler = amp.GradScaler(enabled=enable_amp)

    num_steps = len(data_loader)
    batch_time = AverageMeter()
    loss_meter = AverageMeter()
    norm_meter = AverageMeter()

    start = time.time()
    end = time.time()
    for idx, (samples, targets) in enumerate(data_loader):

        if idx == 0:
            image_size = samples.shape
            logger.info('****input_size {}'.format(str(image_size)))

        samples = samples.cuda(non_blocking=True)  # samples = samples.to(rank)
        targets = targets.cuda(non_blocking=True)  # targets = targets.to(rank)

        if mixup_fn is not None:
            samples, targets = mixup_fn(samples, targets)

        with amp.autocast(enabled=enable_amp):
            outputs = model(samples)
            loss = criterion(outputs, targets)

        if config.TRAIN.ACCUMULATION_STEPS > 1:
            loss = loss / config.TRAIN.ACCUMULATION_STEPS
            scaler.scale(loss).backward()
            if config.TRAIN.CLIP_GRAD:
                grad_norm = torch.nn.utils.clip_grad_norm_(model.parameters(), config.TRAIN.CLIP_GRAD)
            else:
                grad_norm = get_grad_norm(model.parameters())
            if (idx + 1) % config.TRAIN.ACCUMULATION_STEPS == 0:
                # optimizer.step()
                scaler.step(optimizer)
                scaler.update()
                optimizer.zero_grad()
                lr_scheduler.step_update(epoch * num_steps + idx)
        else:
            optimizer.zero_grad()
            scaler.scale(loss).backward()
            scaler.unscale_(optimizer)

            if config.TRAIN.CLIP_GRAD:
                grad_norm = torch.nn.utils.clip_grad_norm_(model.parameters(), config.TRAIN.CLIP_GRAD)
            else:
                grad_norm = get_grad_norm(model.parameters())

            # optimizer.step()
            scaler.step(optimizer)
            scaler.update()

            if model_ema is not None:
                model_ema.update(model)

            lr_scheduler.step_update(epoch * num_steps + idx)

        torch.cuda.synchronize()

        loss_meter.update(loss.item(), targets.size(0))
        norm_meter.update(grad_norm)
        batch_time.update(time.time() - end)
        end = time.time()

        if idx % config.PRINT_FREQ == 0:
            lr = optimizer.param_groups[0]['lr']
            memory_used = torch.cuda.max_memory_allocated() / (1024.0 * 1024.0)
            etas = batch_time.avg * (num_steps - idx)
            logger.info(
                f'Train: [{epoch}/{config.TRAIN.EPOCHS}][{idx}/{num_steps}]\t'
                f'eta {datetime.timedelta(seconds=int(etas))} lr {lr:.6f}\t'
                # f'time {batch_time.val:.4f} ({batch_time.avg:.4f})\t'
                f'loss {loss_meter.val:.4f} ({loss_meter.avg:.4f})\t'
                f'grad_norm {norm_meter.val:.4f} ({norm_meter.avg:.4f})\t'
                f'mem {memory_used:.0f}MB')
    epoch_time = time.time() - start
    logger.info(f"EPOCH {epoch} training takes {datetime.timedelta(seconds=int(epoch_time))}")


@torch.no_grad()
def validate(config, data_loader, model, logger, rank, ema_mode=False):
    criterion = torch.nn.CrossEntropyLoss()
    model.eval()

    batch_time = AverageMeter()
    loss_meter = AverageMeter()
    acc1_meter = AverageMeter()
    acc5_meter = AverageMeter()

    end = time.time()
    for idx, (images, target) in enumerate(data_loader):
        images = images.cuda(non_blocking=True)  # images = images.to(rank)
        target = target.cuda(non_blocking=True)  # target = target.to(rank)

        # compute output
        output = model(images)

        # measure accuracy and record loss
        loss = criterion(output, target)
        acc1, acc5 = accuracy(output, target, topk=(1, 5))

        #  print(acc1, acc5)

        acc1 = reduce_tensor(acc1)
        acc5 = reduce_tensor(acc5)
        loss = reduce_tensor(loss)

        loss_meter.update(loss.item(), target.size(0))
        acc1_meter.update(acc1.item(), target.size(0))
        acc5_meter.update(acc5.item(), target.size(0))

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if idx % config.PRINT_FREQ == 0:
            memory_used = torch.cuda.max_memory_allocated() / (1024.0 * 1024.0)
            logger.info(
                f'Test: [{idx}/{len(data_loader)}]\t'
                f'Time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'
                f'Loss {loss_meter.val:.4f} ({loss_meter.avg:.4f})\t'
                f'Acc@1 {acc1_meter.val:.3f} ({acc1_meter.avg:.3f})\t'
                f'Acc@5 {acc5_meter.val:.3f} ({acc5_meter.avg:.3f})\t'
                f'Mem {memory_used:.0f}MB')

    if ema_mode:
        logger.info(f' EMA Acc@1 {acc1_meter.avg:.3f} Acc@5 {acc5_meter.avg:.3f}')
    else:
        logger.info(f' * Acc@1 {acc1_meter.avg:.3f} Acc@5 {acc5_meter.avg:.3f}')

    return acc1_meter.avg, acc5_meter.avg, loss_meter.avg


@torch.no_grad()
def validate_ema(config, data_loader, model, logger, rank, ema_mode=False):
    #  criterion = torch.nn.CrossEntropyLoss()
    model.eval()
    acc1_meter = AverageMeter()
    acc5_meter = AverageMeter()
    for idx, (images, target) in enumerate(data_loader):
        images = images.cuda(non_blocking=True)  # images = images.to(rank)
        target = target.cuda(non_blocking=True)  # target = target.to(rank)

        # compute output
        output = model(images)

        # measure accuracy and record loss
        # loss = criterion(output, target)
        acc1, acc5 = accuracy(output, target, topk=(1, 5))
        acc1 = reduce_tensor(acc1)
        acc5 = reduce_tensor(acc5)

        acc1_meter.update(acc1.item(), target.size(0))
        acc5_meter.update(acc5.item(), target.size(0))

    logger.info(f' EMA Acc@1 {acc1_meter.avg:.3f} Acc@5 {acc5_meter.avg:.3f}')

    return


@torch.no_grad()
def throughput(data_loader, model, logger):
    model.eval()

    for idx, (images, _) in enumerate(data_loader):
        images = images.cuda(non_blocking=True)
        batch_size = images.shape[0]
        for i in range(50):
            model(images)
        torch.cuda.synchronize()
        logger.info(f"throughput averaged with 30 times")
        tic1 = time.time()
        for i in range(30):
            model(images)
        torch.cuda.synchronize()
        tic2 = time.time()
        logger.info(f"batch_size {batch_size} throughput {30 * batch_size / (tic2 - tic1)}")
        return


def distributed_worker(i, main, parser, config, kwargs):
    setattr(parser, "device_id", i)
    if torch.cuda.is_available():
        torch.cuda.set_device(i)

    ddp_rank = getattr(parser, "local_rank", None)
    if ddp_rank is None:  # torch.multiprocessing.spawn
        ddp_rank = kwargs.get('start_rank', 0) + i
        setattr(parser, "local_rank", ddp_rank)

    node_rank = distributed_init(parser)
    setattr(parser, "local_rank", node_rank)
    main(parser, config, **kwargs)


def main_worker(**kwargs):
    config, opts = parse_option()
    # device set-up 1:random_seed  2:set device  3:benchmark 4:set opts.dev.-
    opts = device_setup(opts)

    num_gpus = getattr(opts, "num_gpus", 1)  # auto set

    world_size = getattr(opts, "world_size", -1)  # -1 means world_size = num_gpus
    use_distributed = getattr(opts, "use_ddp", False)

    if num_gpus <= 1:
        use_distributed = False
    setattr(opts, "use_ddp", use_distributed)

    # No of data workers = no of CPUs (if not specified or -1)
    n_cpus = multiprocessing.cpu_count()
    num_workers = getattr(opts, "num_workers", -1)

    if use_distributed:
        if world_size == -1:
            world_size = num_gpus
            setattr(opts, "world_size", world_size)
        elif world_size != num_gpus:
            world_size = num_gpus
            setattr(opts, "world_size", world_size)

        if num_workers == -1 or num_workers is None:
            setattr(opts, "num_workers", n_cpus // world_size)

        start_rank = getattr(opts, "local_ank", 0)
        setattr(opts, "local_rank", None)
        kwargs['start_rank'] = start_rank
        torch.multiprocessing.spawn(
            fn=distributed_worker,
            args=(main, opts, config, kwargs),
            nprocs=num_gpus,
        )
    else:
        if num_workers == -1:
            setattr(opts, "num_workers", n_cpus)

        setattr(opts, "device_id", None)
        main(opts=opts, config=config, **kwargs)


if __name__ == "__main__":
    main_worker()
