# --------------------------------------------------------
# Swin Transformer
# Copyright (c) 2021 Microsoft
# Licensed under The MIT License [see LICENSE for details]
# Written by Ze Liu
# --------------------------------------------------------
import socket
import os
import random
from logging import warning

import numpy as np
import torch
import torch.distributed as dist
from timm.utils import get_state_dict
from torch.cuda import amp


# try:
#     # noinspection PyUnresolvedReferences
#     from apex import amp
# except ImportError:
#     amp = None


def load_checkpoint(config, model, optimizer, lr_scheduler, logger, model_ema=None):
    logger.info(f"==============> Resuming form {config.MODEL.RESUME}....................")
    if config.MODEL.RESUME.startswith('https'):
        checkpoint = torch.hub.load_state_dict_from_url(
            config.MODEL.RESUME, map_location='cpu', check_hash=True)
    else:
        checkpoint = torch.load(config.MODEL.RESUME, map_location='cpu')

    msg = model.load_state_dict(checkpoint['model'], strict=False)
    logger.info(msg)

    if config.EMA.USE:
        if 'model_ema' in checkpoint.keys():
            print("load ema model")
            msg = model_ema.ema.load_state_dict(checkpoint['model_ema'], strict=False)
        else:
            msg = model_ema.ema.load_state_dict(checkpoint['model'], strict=False)
        logger.info(msg)

    max_accuracy = 0.0
    if not config.EVAL_MODE and 'optimizer' in checkpoint and 'lr_scheduler' in checkpoint and 'epoch' in checkpoint:
        optimizer.load_state_dict(checkpoint['optimizer'])
        lr_scheduler.load_state_dict(checkpoint['lr_scheduler'])
        config.defrost()
        config.TRAIN.START_EPOCH = checkpoint['epoch'] + 1
        config.freeze()
        # if 'amp' in checkpoint and config.AMP_OPT_LEVEL != "O0" and checkpoint['config'].AMP_OPT_LEVEL != "O0":
        if 'amp' in checkpoint:
            amp.load_state_dict(checkpoint['amp'])
        logger.info(f"=> loaded successfully '{config.MODEL.RESUME}' (epoch {checkpoint['epoch']})")
        if 'max_accuracy' in checkpoint:
            max_accuracy = checkpoint['max_accuracy']

    del checkpoint
    torch.cuda.empty_cache()
    return max_accuracy


def save_checkpoint(config, epoch, model, max_accuracy, optimizer, lr_scheduler, logger, model_ema):
    save_state = {'model': model.state_dict(),
                  'optimizer': optimizer.state_dict(),
                  'lr_scheduler': lr_scheduler.state_dict(),
                  'max_accuracy': max_accuracy,
                  'epoch': epoch,
                  'config': config}

    # if config.AMP_OPT_LEVEL != "O0":
    #     save_state['amp'] = amp.state_dict()

    if model_ema is not None:
        # save_state['model_ema'] = get_state_dict(model_ema)
        save_state['model_ema'] = get_state_dict(model_ema)

    save_path = os.path.join(config.OUTPUT, f'ckpt_epoch_{epoch}.pth')
    logger.info(f"{save_path} saving......")
    torch.save(save_state, save_path)
    logger.info(f"{save_path} saved !!!")


def get_grad_norm(parameters, norm_type=2):
    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]
    parameters = list(filter(lambda p: p.grad is not None, parameters))
    norm_type = float(norm_type)
    total_norm = 0
    for p in parameters:
        param_norm = p.grad.data.norm(norm_type)
        total_norm += param_norm.item() ** norm_type
    total_norm = total_norm ** (1. / norm_type)
    return total_norm


def auto_resume_helper(output_dir):
    checkpoints = os.listdir(output_dir)
    checkpoints = [ckpt for ckpt in checkpoints if ckpt.endswith('pth')]
    print(f"All checkpoints founded in {output_dir}: {checkpoints}")
    if len(checkpoints) > 0:
        latest_checkpoint = max([os.path.join(output_dir, d) for d in checkpoints], key=os.path.getmtime)
        print(f"The latest checkpoint founded: {latest_checkpoint}")
        resume_file = latest_checkpoint
    else:
        resume_file = None
    return resume_file


def reduce_tensor(tensor):
    rt = tensor.clone()
    dist.all_reduce(rt, op=dist.ReduceOp.SUM)
    rt /= dist.get_world_size()
    return rt


def distributed_init(opts) -> int:
    ddp_url = getattr(opts, "dist_url", None)
    ddp_port = getattr(opts, "dist_port", 6006)

    if ddp_url is None:
        hostname = socket.gethostname()
        ddp_url = 'tcp://{}:{}'.format(hostname, ddp_port)
        setattr(opts, "dist_url", ddp_url)

    node_rank = getattr(opts, "local_rank", 0)
    world_size = getattr(opts, "world_size", 0)
    if torch.distributed.is_initialized():
        warning('DDP is already initialized and cannot be initialize twice!')
    else:

        dist_backend = "gloo"
        if dist.is_nccl_available():
            dist_backend = 'nccl'

        dist.init_process_group(
            backend=dist_backend,
            init_method=ddp_url,
            world_size=world_size,
            rank=node_rank
        )

        # perform a dummy all-reduce to initialize the NCCL communicator
        if torch.cuda.is_available():
            dist.all_reduce(torch.zeros(1).cuda())

    node_rank = torch.distributed.get_rank()
    setattr(opts, "local_rank", node_rank)
    return node_rank


def device_setup(opts):
    random_seed = getattr(opts, "random_seed", 0)
    random.seed(random_seed)
    torch.manual_seed(random_seed)
    np.random.seed(random_seed)

    n_gpus = torch.cuda.device_count()
    if n_gpus == 0:
        device = torch.device('cpu')
        n_gpus = 0
    else:
        device = torch.device('cuda')

        if torch.backends.cudnn.is_available():
            import torch.backends.cudnn as cudnn
            torch.backends.cudnn.enabled = True
            cudnn.benchmark = False
            cudnn.deterministic = True

    setattr(opts, "device", device)
    setattr(opts, "num_gpus", n_gpus)

    return opts
